# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-02-28 15:53
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("uk_geo_utils", "0003_auto_20171030_1504")]

    operations = [
        migrations.RenameField(
            model_name="onspd", old_name="hro", new_name="nhser"
        ),
        migrations.RenameField(
            model_name="onspd", old_name="gor", new_name="rgn"
        ),
        migrations.RemoveField(model_name="onspd", name="cened"),
        migrations.RemoveField(model_name="onspd", name="edind"),
        migrations.RemoveField(model_name="onspd", name="lea"),
        migrations.RemoveField(model_name="onspd", name="oldha"),
        migrations.RemoveField(model_name="onspd", name="oldpct"),
        migrations.RemoveField(model_name="onspd", name="oshaprev"),
        migrations.RemoveField(model_name="onspd", name="psed"),
        migrations.RemoveField(model_name="onspd", name="streg"),
        migrations.RemoveField(model_name="onspd", name="ward98"),
        migrations.RemoveField(model_name="onspd", name="wardc91"),
        migrations.RemoveField(model_name="onspd", name="wardo91"),
        migrations.AddField(
            model_name="onspd",
            name="ced",
            field=models.CharField(blank=True, max_length=9),
        ),
    ]



import abc

from django.core.exceptions import ObjectDoesNotExist

from uk_geo_utils.helpers import (
    AddressSorter,
    Postcode,
    get_address_model,
    get_onspd_model,
    get_onsud_model,
)


class AddressBaseException(Exception):
    pass


class CodesNotFoundException(AddressBaseException):
    pass


class MultipleCodesException(AddressBaseException):
    pass


class StrictMatchException(AddressBaseException):
    pass


class NorthernIrelandException(ObjectDoesNotExist):
    pass


class AddressBaseNotImportedException(ObjectDoesNotExist):
    pass


class OnsudNotImportedException(ObjectDoesNotExist):
    pass


class OnspdNotImportedException(ObjectDoesNotExist):
    pass


class BaseGeocoder(metaclass=abc.ABCMeta):
    def __init__(self, postcode):
        self.postcode = Postcode(postcode)

    @property
    @abc.abstractmethod
    def centroid(self):
        pass

    @abc.abstractmethod
    def get_code(code_type, uprn=None):
        pass


class AddressBaseGeocoder(BaseGeocoder):
    def __init__(self, postcode):
        self.postcode = Postcode(postcode)
        if self.postcode.territory == "NI":
            raise NorthernIrelandException("Postcode is in Northern Ireland")

        self.onsud_model = get_onsud_model()
        self.address_model = get_address_model()

        # check the data we need exists
        if not self.address_model.objects.all().exists():
            raise AddressBaseNotImportedException("Address Base table is empty")
        if not self.onsud_model.objects.all().exists():
            raise OnsudNotImportedException("ONSUD table is empty")

        self._addresses = self.address_model.objects.filter(
            postcode=self.postcode.with_space
        ).order_by("uprn")
        if not self._addresses:
            raise self.address_model.DoesNotExist(
                "No addresses found for postcode %s" % (self.postcode)
            )

        self._uprns = self.onsud_model.objects.filter(
            uprn__in=self.uprns
        ).order_by("uprn")

    @property
    def uprns(self):
        return [a.uprn for a in self._addresses]

    @property
    def centroid(self):
        type_d_addresses = self._addresses.filter(addressbase_postal="D")
        if len(type_d_addresses) > 0:
            return type_d_addresses.centroid
        return self._addresses.centroid

    @property
    def addresses(self):
        sorter = AddressSorter(self._addresses)
        return sorter.natural_sort()

    def get_point(self, uprn):
        return self._addresses.get_cached(uprn).location

    def get_code(self, code_type, uprn=None, strict=False):
        # check the code_type field exists on our model
        code_type_field = self.onsud_model._meta.get_field(code_type)

        if uprn:
            self._addresses.get_cached(uprn)
            return getattr(self._uprns.get_cached(uprn), code_type)

        if len(self._uprns) == 0:
            # No records in the ONSUD table were found for the given UPRNs
            # because...reasons
            raise CodesNotFoundException(
                "Found no records in ONSUD for supplied UPRNs"
            )
        if len(self._addresses) != len(self._uprns):
            if strict:
                for uprn in self.uprns:
                    try:
                        self._uprns.get_cached(uprn)
                    except self.onsud_model.DoesNotExist:
                        raise StrictMatchException(
                            "expected UPRN %s not found in ONSUD" % uprn
                        )
            else:
                # if not strict, ignore this condition
                pass

        codes = {getattr(u, code_type_field.attname) for u in self._uprns}
        if len(codes) == 1:
            # all the uprns supplied are in the same area
            return list(codes)[0]
        raise MultipleCodesException(
            "Postcode %s covers UPRNs in more than one '%s' area"
            % (self.postcode, code_type)
        )


class OnspdGeocoder(BaseGeocoder):
    def __init__(self, postcode):
        self.postcode = Postcode(postcode)
        self.onspd_model = get_onspd_model()

        if not self.onspd_model.objects.all().exists():
            raise OnspdNotImportedException("ONSPD table is empty")

        self.record = self.onspd_model.objects.get(
            pcds=self.postcode.with_space, doterm=""
        )

    @property
    def centroid(self):
        return self.record.location

    def get_code(self, code_type):
        return getattr(self.record, code_type)



import os

from django.core.management.base import BaseCommand
from django.db import connection, transaction

from uk_geo_utils.helpers import get_address_model


class Command(BaseCommand):
    help = (
        "Deletes all data in Address model AND any related tables,"
        "and replaces Address model data with that in the cleaned AddressBase CSVs."
        "Data in related tables will need to be imported/rebuilt seperately"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "cleaned_ab_path",
            help="The path to the folder containing the cleaned AddressBase CSVs",
        )
        parser.add_argument(
            "-t",
            "--transaction",
            help="Run the import in a transaction",
            action="store_true",
            default=False,
            dest="transaction",
        )

    def handle(self, *args, **kwargs):
        self.table_name = get_address_model()._meta.db_table
        self.path = kwargs["cleaned_ab_path"]
        if kwargs["transaction"]:
            with transaction.atomic():
                self.import_addressbase()
        else:
            self.import_addressbase()

    def import_addressbase(self):
        cleaned_file_path = os.path.abspath(
            os.path.join(self.path, "addressbase_cleaned.csv")
        )

        with open(cleaned_file_path, "r") as fp:
            cursor = connection.cursor()
            self.stdout.write("clearing existing data..")
            cursor.execute("TRUNCATE TABLE %s CASCADE;" % (self.table_name))

            self.stdout.write("importing from %s.." % (cleaned_file_path))
            cursor.copy_expert(
                """
                COPY %s (UPRN,address,postcode,location,addressbase_postal)
                FROM STDIN (FORMAT CSV, DELIMITER ',', quote '"');
            """
                % (self.table_name),
                fp,
            )

        self.stdout.write("...done")



import csv
import glob
import os

from django.core.management.base import BaseCommand

from uk_geo_utils.helpers import PAFAddressFormatter


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "ab_path",
            help="The path to the folder containing the AddressBase CSVs",
        )

    def handle(self, *args, **kwargs):
        self.fieldnames = [
            "UPRN",
            "OS_ADDRESS_TOID",
            "UDPRN",
            "ORGANISATION_NAME",
            "DEPARTMENT_NAME",
            "PO_BOX_NUMBER",
            "SUB_BUILDING_NAME",
            "BUILDING_NAME",
            "BUILDING_NUMBER",
            "DEPENDENT_THOROUGHFARE",
            "THOROUGHFARE",
            "POST_TOWN",
            "DOUBLE_DEPENDENT_LOCALITY",
            "DEPENDENT_LOCALITY",
            "POSTCODE",
            "POSTCODE_TYPE",
            "X_COORDINATE",
            "Y_COORDINATE",
            "LATITUDE",
            "LONGITUDE",
            "RPC",
            "COUNTRY",
            "CHANGE_TYPE",
            "LA_START_DATE",
            "RM_START_DATE",
            "LAST_UPDATE_DATE",
            "CLASS",
        ]
        self.base_path = os.path.abspath(kwargs["ab_path"])
        out_path = os.path.join(self.base_path, "addressbase_cleaned.csv")

        files = glob.glob(os.path.join(self.base_path, "*.csv"))
        if not files:
            raise FileNotFoundError(
                "No CSV files found in %s" % (self.base_path)
            )

        with open(out_path, "w") as out_file:
            for csv_path in files:
                if csv_path.endswith("cleaned.csv"):
                    continue
                self.out_csv = csv.DictWriter(
                    out_file,
                    fieldnames=[
                        "UPRN",
                        "address",
                        "postcode",
                        "location",
                        "addressbase_postal",
                    ],
                )
                self.stdout.write(csv_path)
                self.clean_csv(csv_path)
                out_file.flush()

    def line_filter(self, csv_path):
        with open(csv_path) as csv_file:
            for line in csv.DictReader(csv_file, fieldnames=self.fieldnames):
                # Do any filtering we might need to do here
                yield line

    def clean_csv(self, csv_path):
        for line in self.line_filter(csv_path):
            self.out_csv.writerow(self.clean_output_line(line))

    def clean_address(self, line):
        address_fields = [
            "ORGANISATION_NAME",
            "DEPARTMENT_NAME",
            "PO_BOX_NUMBER",
            "SUB_BUILDING_NAME",
            "BUILDING_NAME",
            "BUILDING_NUMBER",
            "DEPENDENT_THOROUGHFARE",
            "THOROUGHFARE",
            "DOUBLE_DEPENDENT_LOCALITY",
            "DEPENDENT_LOCALITY",
            "POST_TOWN",
        ]
        kwargs = {k.lower(): line[k] for k in line if k in address_fields}
        return PAFAddressFormatter(**kwargs).generate_address_label()

    def clean_output_line(self, line):
        data = {}
        data["UPRN"] = line["UPRN"]
        data["address"] = self.clean_address(line)
        data["postcode"] = line["POSTCODE"]
        data["location"] = "SRID=4326;POINT({} {})".format(
            line["LONGITUDE"], line["LATITUDE"]
        )
        data["addressbase_postal"] = "D"
        return data



import csv
import glob
import os

from django.core.management.base import BaseCommand

from uk_geo_utils.helpers import LocalAuthAddressFormatter, PAFAddressFormatter


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "ab_path",
            help="The path to the folder containing the AddressBase CSVs",
        )

    def handle(self, *args, **kwargs):
        self.fieldnames = [
            "UPRN",
            "UDPRN",
            "CHANGE_TYPE",
            "STATE",
            "STATE_DATE",
            "CLASS",
            "PARENT_UPRN",
            "X_COORDINATE",
            "Y_COORDINATE",
            "LATITUDE",
            "LONGITUDE",
            "RPC",
            "LOCAL_CUSTODIAN_CODE",
            "COUNTRY",
            "LA_START_DATE",
            "LAST_UPDATE_DATE",
            "ENTRY_DATE",
            "RM_ORGANISATION_NAME",
            "LA_ORGANISATION",
            "DEPARTMENT_NAME",
            "LEGAL_NAME",
            "SUB_BUILDING_NAME",
            "BUILDING_NAME",
            "BUILDING_NUMBER",
            "SAO_START_NUMBER",
            "SAO_START_SUFFIX",
            "SAO_END_NUMBER",
            "SAO_END_SUFFIX",
            "SAO_TEXT",
            "ALT_LANGUAGE_SAO_TEXT",
            "PAO_START_NUMBER",
            "PAO_START_SUFFIX",
            "PAO_END_NUMBER",
            "PAO_END_SUFFIX",
            "PAO_TEXT",
            "ALT_LANGUAGE_PAO_TEXT",
            "USRN",
            "USRN_MATCH_INDICATOR",
            "AREA_NAME",
            "LEVEL",
            "OFFICIAL_FLAG",
            "OS_ADDRESS_TOID",
            "OS_ADDRESS_TOID_VERSION",
            "OS_ROADLINK_TOID",
            "OS_ROADLINK_TOID_VERSION",
            "OS_TOPO_TOID",
            "OS_TOPO_TOID_VERSION",
            "VOA_CT_RECORD",
            "VOA_NDR_RECORD",
            "STREET_DESCRIPTION",
            "ALT_LANGUAGE_STREET_DESCRIPTION",
            "DEPENDENT_THOROUGHFARE",
            "THOROUGHFARE",
            "WELSH_DEPENDENT_THOROUGHFARE",
            "WELSH_THOROUGHFARE",
            "DOUBLE_DEPENDENT_LOCALITY",
            "DEPENDENT_LOCALITY",
            "LOCALITY",
            "WELSH_DEPENDENT_LOCALITY",
            "WELSH_DOUBLE_DEPENDENT_LOCALITY",
            "TOWN_NAME",
            "ADMINISTRATIVE_AREA",
            "POST_TOWN",
            "WELSH_POST_TOWN",
            "POSTCODE",
            "POSTCODE_LOCATOR",
            "POSTCODE_TYPE",
            "DELIVERY_POINT_SUFFIX",
            "ADDRESSBASE_POSTAL",
            "PO_BOX_NUMBER",
            "WARD_CODE",
            "PARISH_CODE",
            "RM_START_DATE",
            "MULTI_OCC_COUNT",
            "VOA_NDR_P_DESC_CODE",
            "VOA_NDR_SCAT_CODE",
            "ALT_LANGUAGE",
        ]
        self.base_path = os.path.abspath(kwargs["ab_path"])
        out_path = os.path.join(self.base_path, "addressbase_cleaned.csv")

        files = glob.glob(os.path.join(self.base_path, "*.csv"))
        if not files:
            raise FileNotFoundError(
                "No CSV files found in %s" % (self.base_path)
            )

        with open(out_path, "w") as out_file:
            for csv_path in files:
                if csv_path.endswith("cleaned.csv"):
                    continue
                self.out_csv = csv.DictWriter(
                    out_file,
                    fieldnames=[
                        "UPRN",
                        "address",
                        "postcode",
                        "location",
                        "addressbase_postal",
                    ],
                )
                self.stdout.write(csv_path)
                self.clean_csv(csv_path)
                out_file.flush()

    def line_filter(self, csv_path):
        with open(csv_path) as csv_file:
            for line in csv.DictReader(csv_file, fieldnames=self.fieldnames):
                """
                - Get rid of type N UPRNs (but keep type C, D and L)
                - Get rid of records in the Isle of Man/Channel Islands
                  (but keep England, Wales, Scotland and NI)
                """
                if line["ADDRESSBASE_POSTAL"] != "N" and line["COUNTRY"] in [
                    "E",
                    "W",
                    "S",
                    "N",
                ]:
                    yield line

    def clean_csv(self, csv_path):
        for line in self.line_filter(csv_path):
            self.out_csv.writerow(self.clean_output_line(line))

    def clean_address(self, line):
        if line["ADDRESSBASE_POSTAL"] == "D":
            address_fields = [
                "DEPARTMENT_NAME",
                "PO_BOX_NUMBER",
                "SUB_BUILDING_NAME",
                "BUILDING_NAME",
                "BUILDING_NUMBER",
                "DEPENDENT_THOROUGHFARE",
                "THOROUGHFARE",
                "DOUBLE_DEPENDENT_LOCALITY",
                "DEPENDENT_LOCALITY",
                "POST_TOWN",
            ]
            kwargs = {k.lower(): line[k] for k in line if k in address_fields}
            kwargs["organisation_name"] = line["RM_ORGANISATION_NAME"]
            return PAFAddressFormatter(**kwargs).generate_address_label()
        address_fields = [
            "SAO_START_NUMBER",
            "SAO_START_SUFFIX",
            "SAO_END_NUMBER",
            "SAO_END_SUFFIX",
            "SAO_TEXT",
            "PAO_START_NUMBER",
            "PAO_START_SUFFIX",
            "PAO_END_NUMBER",
            "PAO_END_SUFFIX",
            "PAO_TEXT",
            "STREET_DESCRIPTION",
            "LOCALITY",
            "TOWN_NAME",
        ]
        kwargs = {k.lower(): line[k] for k in line if k in address_fields}
        kwargs["organisation_name"] = line["LA_ORGANISATION"]
        return LocalAuthAddressFormatter(**kwargs).generate_address_label()

    def clean_output_line(self, line):
        data = {}
        data["UPRN"] = line["UPRN"]
        data["address"] = self.clean_address(line)
        if line["ADDRESSBASE_POSTAL"] == "D":
            data["postcode"] = line["POSTCODE"]
        else:
            data["postcode"] = line["POSTCODE_LOCATOR"]
        data["location"] = "SRID=4326;POINT({} {})".format(
            line["LONGITUDE"], line["LATITUDE"]
        )
        data["addressbase_postal"] = line["ADDRESSBASE_POSTAL"]
        return data



# Generated by Django 4.1.3 on 2022-11-01 16:10

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("uk_geo_utils", "0008_address_addressbase_postal"),
    ]

    operations = [
        migrations.AddField(
            model_name="onspd",
            name="lsoa21",
            field=models.CharField(max_length=9, null=True),
        ),
        migrations.AddField(
            model_name="onspd",
            name="msoa21",
            field=models.CharField(max_length=9, null=True),
        ),
        migrations.AddField(
            model_name="onspd",
            name="oa21",
            field=models.CharField(max_length=9, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-06-12 16:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("uk_geo_utils", "0005_onsud_ced")]

    operations = [
        migrations.AddField(
            model_name="onspd",
            name="calncv",
            field=models.CharField(blank=True, max_length=9),
        ),
        migrations.AddField(
            model_name="onspd",
            name="stp",
            field=models.CharField(blank=True, max_length=9),
        ),
        migrations.AddField(
            model_name="onspd",
            name="streg",
            field=models.CharField(blank=True, max_length=1),
        ),
    ]



import glob
import os

from django.core.management.base import BaseCommand
from django.db import connection, transaction

from uk_geo_utils.helpers import get_onspd_model

HEADERS = {
    "may2018": """
        pcd, pcd2, pcds, dointr, doterm, oscty, ced, oslaua, osward,
        parish, usertype, oseast1m, osnrth1m, osgrdind, oshlthau,
        nhser, ctry, rgn, streg, pcon, eer, teclec, ttwa, pct, nuts,
        statsward, oa01, casward, park, lsoa01, msoa01, ur01ind,
        oac01, oa11, lsoa11, msoa11, wz11, ccg, bua11, buasd11,
        ru11ind, oac11, lat, long, lep1, lep2, pfa, imd, calncv, stp
        """,
    "aug2022": """
        pcd, pcd2, pcds, dointr, doterm, oscty, ced, oslaua, osward,
        parish, usertype, oseast1m, osnrth1m, osgrdind, oshlthau,
        nhser, ctry, rgn, streg, pcon, eer, teclec, ttwa, pct, nuts,
        statsward, oa01, casward, park, lsoa01, msoa01, ur01ind,
        oac01, oa11, lsoa11, msoa11, wz11, ccg, bua11, buasd11,
        ru11ind, oac11, lat, long, lep1, lep2, pfa, imd, calncv, stp,
        oa21, lsoa21, msoa21
        """,
}


class Command(BaseCommand):
    """
    To import ONSPD, grab the latest release:
    https://ons.maps.arcgis.com/home/search.html?t=content&q=ONS%20Postcode%20Directory
    and run
    python manage.py import_onspd /path/to/data
    """

    def add_arguments(self, parser):
        parser.add_argument(
            "path", help="Path to the directory containing the ONSPD CSVs"
        )
        parser.add_argument(
            "-t",
            "--transaction",
            help="Run the import in a transaction",
            action="store_true",
            default=False,
            dest="transaction",
        )

        parser.add_argument(
            "--header",
            help="Specify which header the csv has",
            default="aug2022",
            choices=["may2018", "aug2022"],
        )

    def handle(self, *args, **kwargs):
        self.table_name = get_onspd_model()._meta.db_table
        self.path = kwargs["path"]
        self.header = HEADERS[kwargs.get("header", "aug2022")]
        if kwargs["transaction"]:
            with transaction.atomic():
                self.import_onspd()
        else:
            self.import_onspd()

    def import_onspd(self):
        glob_str = os.path.join(self.path, "*.csv")
        files = glob.glob(glob_str)
        if not files:
            raise FileNotFoundError("No CSV files found in %s" % (self.path))

        cursor = connection.cursor()
        self.stdout.write("clearing existing data..")
        cursor.execute("TRUNCATE TABLE %s;" % (self.table_name))

        self.stdout.write("importing from files..")
        for f in files:
            self.stdout.write(f)
            with open(f, "r") as fp:
                cursor.copy_expert(
                    """
                    COPY %s (
                    %s
                    ) FROM STDIN (FORMAT CSV, DELIMITER ',', quote '"', HEADER);
                """
                    % (self.table_name, self.header),
                    fp,
                )

        # turn text lng/lat into a Point() field
        cursor.execute(
            """
            UPDATE %s SET location=CASE
                WHEN ("long"='0.000000' AND lat='99.999999')
                THEN NULL
                ELSE ST_GeomFromText('POINT(' || "long" || ' ' || lat || ')',4326)
            END
        """
            % (self.table_name)
        )

        self.stdout.write("...done")



import os

from setuptools import find_packages, setup

import uk_geo_utils

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


def _get_description():
    try:
        path = os.path.join(os.path.dirname(__file__), "README.md")
        with open(path, encoding="utf-8") as f:
            return f.read()
    except IOError:
        return ""


def get_version():
    return uk_geo_utils.__version__


setup(
    name="uk_geo_utils",
    version=get_version(),
    author="chris48s",
    license="MIT",
    url="https://github.com/DemocracyClub/uk-geo-utils",
    packages=find_packages(),
    include_package_data=True,
    description="Django app for working with OS Addressbase, ONSUD and ONSPD",
    long_description=_get_description(),
    long_description_content_type="text/markdown",
    install_requires=["Django>=4.2", "psycopg2-binary"],
    extras_require={"development": ["coveralls", "mkdocs", "ruff==0.6.2"]},
    classifiers=[
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.1",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.12",
    ],
    project_urls={
        "Documentation": "https://democracyclub.github.io/uk-geo-utils/",
        "Source": "https://github.com/DemocracyClub/uk-geo-utils",
    },
)



# Generated by Django 2.0.8 on 2020-01-31 15:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("uk_geo_utils", "0007_remove_onsud_ctry_flag")]

    operations = [
        migrations.AddField(
            model_name="address",
            name="addressbase_postal",
            field=models.CharField(default="D", max_length=1),
            preserve_default=False,
        )
    ]
